from .adc_base import ADCBase


class LightSensor(ADCBase):
    """Encapsulates the behaviour of a light sensor module.

    A simple analogue photo transistor is used to detect the intensity of the light striking
    the sensor. The component contains a photoresistor which detects light intensity. The
    resistance decreases as light intensity increases; thus the brighter the light, the
    higher the voltage.

    Uses an Analog-to-Digital Converter (ADC) to turn the analog reading from the sensor
    into a digital value.

    :param str port_name: The ID for the port to which this component is connected
    """

    def __init__(self, port_name, pin_number=1, name="light_sensor"):
        ADCBase.__init__(self, port_name=port_name, pin_number=pin_number, name=name)

    @property
    def reading(self):
        """Take a reading from the sensor.

        :return: A value representing the amount of light striking the sensor at the current time
            from 0 to 999.
        :rtype: float
        """
        return int(self.read(number_of_samples=3))

    @property
    def value(self):
        """Get a simple binary value based on a reading from the device.

        :return: 1 if the sensor is detecting any light, 0 otherwise
        :rtype: integer
        """
        if self.reading > 0:
            return 1
        else:
            return 0



import numpy as np

from pitop.common.firmware_device import FirmwareDevice, FirmwareDeviceID

from .common.imu_registers import (
    ImuRegisters,
    MagCalHardTypes,
    MagCalRegisterTypes,
    MagCalSoftTypes,
    OrientationRegisterTypes,
    RawRegisterTypes,
    RegisterTypes,
    ScaleMappings,
)
from .plate_interface import PlateInterface


class ImuController:
    """Class used to read/write IMU registers from the Expansion Plate MCU."""

    __ORIENTATION_DATA_SCALE = 100.0
    __MAG_SIGNED_RANGE = 4900.0
    __16BIT_SIGNED_RANGE = 2**15
    __HARD_IRON_SCALE_FACTOR = 10.0
    __SOFT_IRON_SCALE_FACTOR = 1000.0

    def __init__(self):
        device = FirmwareDevice(FirmwareDeviceID.pt4_expansion_plate)
        if device.get_sch_hardware_version_major() == 4:
            raise Exception("This Expansion Plate doesn't have an IMU")

        self.__data_registers = ImuRegisters.DATA
        self.__enable_registers = ImuRegisters.ENABLE
        self.__config_registers = ImuRegisters.CONFIG
        self.__mag_cal_registers = ImuRegisters.MAGCAL
        self.__mcu_device = PlateInterface().get_device_mcu()
        self.__acc_enable = False
        self.__gyro_enable = False
        self.__mag_enable = False
        self.__orientation_enable = False
        self.__acc_scaler = None
        self.__gyro_scaler = None

        # TODO - review member variables vs get/set properties
        self.acc_scaler = 2
        self.gyro_scaler = 250

    def cleanup(self):
        self.acc_enable = False
        self.gyro_enable = False
        self.mag_enable = False
        self.orientation_enable = False

    def __set_imu_config(self, imu_function: int, enable: bool):
        self.__mcu_device.write_word(
            self.__enable_registers[imu_function],
            int(enable),
            little_endian=True,
            signed=False,
        )

    @property
    def acc_enable(self):
        return self.__acc_enable

    @acc_enable.setter
    def acc_enable(self, enable: bool):
        self.__set_imu_config(RegisterTypes.ACC, enable)
        self.__acc_enable = enable

    @property
    def gyro_enable(self):
        return self.__gyro_enable

    @gyro_enable.setter
    def gyro_enable(self, enable: bool):
        self.__set_imu_config(RegisterTypes.GYRO, enable)
        self.__gyro_enable = enable

    @property
    def mag_enable(self):
        return self.__mag_enable

    @mag_enable.setter
    def mag_enable(self, enable: bool):
        self.__set_imu_config(RegisterTypes.MAG, enable)
        self.__mag_enable = enable

    @property
    def orientation_enable(self):
        return self.__orientation_enable

    @orientation_enable.setter
    def orientation_enable(self, enable: bool):
        self.__set_imu_config(RegisterTypes.ORIENTATION, enable)
        self.__orientation_enable = enable

    def __data_scale_config_write(self, register_type: int, scaler: int):
        if scaler not in ScaleMappings[register_type].keys():
            raise ValueError(
                f"Valid values for scalers are: {list(ScaleMappings[register_type].keys())}"
            )
        byte = ScaleMappings[register_type][scaler]
        self.__mcu_device.write_byte(self.__config_registers[register_type], byte)

    def __data_scale_config_read(self, register_type: int):
        byte = self.__mcu_device.read_unsigned_byte(
            self.__config_registers[register_type]
        )
        scaler = next(
            key for key, value in ScaleMappings[register_type].items() if value == byte
        )
        return scaler

    @property
    def acc_scaler(self):
        return self.__data_scale_config_read(RegisterTypes.ACC)

    @acc_scaler.setter
    def acc_scaler(self, scaler: int):
        self.__acc_scaler = scaler
        self.__data_scale_config_write(RegisterTypes.ACC, scaler)

    @property
    def gyro_scaler(self):
        return self.__data_scale_config_read(RegisterTypes.GYRO)

    @gyro_scaler.setter
    def gyro_scaler(self, scaler: int):
        self.__gyro_scaler = scaler
        self.__data_scale_config_write(RegisterTypes.GYRO, scaler)

    @property
    def accelerometer_raw(self):
        if not self.acc_enable:
            self.acc_enable = True

        imu_acc_raw = self.__get_raw_data(RegisterTypes.ACC)

        imu_acc_scaled = tuple(
            [
                axis / (self.__16BIT_SIGNED_RANGE / float(self.__acc_scaler))
                for axis in imu_acc_raw
            ]
        )

        return imu_acc_scaled

    @property
    def gyroscope_raw(self):
        if not self.gyro_enable:
            self.gyro_enable = True

        imu_gyro_raw = self.__get_raw_data(RegisterTypes.GYRO)

        imu_gyro_scaled = tuple(
            [
                axis / (self.__16BIT_SIGNED_RANGE / float(self.__gyro_scaler))
                for axis in imu_gyro_raw
            ]
        )

        return imu_gyro_scaled

    @property
    def magnetometer_raw(self):
        if not self.mag_enable:
            self.mag_enable = True

        mag_raw = self.__get_raw_data(RegisterTypes.MAG)

        mag_raw_scaled = tuple(
            axis * self.__MAG_SIGNED_RANGE / self.__16BIT_SIGNED_RANGE
            for axis in mag_raw
        )

        return mag_raw_scaled

    @property
    def orientation_data(self):
        if not self.orientation_enable:
            self.orientation_enable = True

        roll = (
            self.__read_imu_data(
                self.__data_registers[RegisterTypes.ORIENTATION][
                    OrientationRegisterTypes.ROLL
                ]
            )
            / self.__ORIENTATION_DATA_SCALE
        )
        pitch = (
            self.__read_imu_data(
                self.__data_registers[RegisterTypes.ORIENTATION][
                    OrientationRegisterTypes.PITCH
                ]
            )
            / self.__ORIENTATION_DATA_SCALE
        )
        yaw = (
            self.__read_imu_data(
                self.__data_registers[RegisterTypes.ORIENTATION][
                    OrientationRegisterTypes.YAW
                ]
            )
            / self.__ORIENTATION_DATA_SCALE
        )

        return roll, pitch, yaw

    def read_mag_cal_params(self):
        x, y, z = self.__read_mag_cal_hard()
        hard_iron_offset = np.array([[x, y, z]]).T / self.__HARD_IRON_SCALE_FACTOR
        xx, yy, zz, xy, xz, yz = self.__read_mag_cal_soft()
        soft_iron_matrix = (
            np.array([[xx, xy, xz], [xy, yy, yz], [xz, yz, zz]])
            / self.__SOFT_IRON_SCALE_FACTOR
        )
        return hard_iron_offset, soft_iron_matrix

    def write_mag_cal_params(self, hard_iron_offset, soft_iron_matrix):
        hard_x = int(round(hard_iron_offset[0][0]))
        hard_y = int(round(hard_iron_offset[1][0]))
        hard_z = int(round(hard_iron_offset[2][0]))

        self.__write_mag_cal(hard_x, MagCalRegisterTypes.HARD, MagCalHardTypes.X)
        self.__write_mag_cal(hard_y, MagCalRegisterTypes.HARD, MagCalHardTypes.Y)
        self.__write_mag_cal(hard_z, MagCalRegisterTypes.HARD, MagCalHardTypes.Z)

        # matrix is symmetric so only need 6 values
        soft_xx = int(round(soft_iron_matrix[0][0] * self.__SOFT_IRON_SCALE_FACTOR))
        soft_xy = int(round(soft_iron_matrix[0][1] * self.__SOFT_IRON_SCALE_FACTOR))
        soft_xz = int(round(soft_iron_matrix[0][2] * self.__SOFT_IRON_SCALE_FACTOR))
        soft_yy = int(round(soft_iron_matrix[1][1] * self.__SOFT_IRON_SCALE_FACTOR))
        soft_yz = int(round(soft_iron_matrix[1][2] * self.__SOFT_IRON_SCALE_FACTOR))
        soft_zz = int(round(soft_iron_matrix[2][2] * self.__SOFT_IRON_SCALE_FACTOR))

        self.__write_mag_cal(soft_xx, MagCalRegisterTypes.SOFT, MagCalSoftTypes.XX)
        self.__write_mag_cal(soft_yy, MagCalRegisterTypes.SOFT, MagCalSoftTypes.YY)
        self.__write_mag_cal(soft_zz, MagCalRegisterTypes.SOFT, MagCalSoftTypes.ZZ)
        self.__write_mag_cal(soft_xy, MagCalRegisterTypes.SOFT, MagCalSoftTypes.XY)
        self.__write_mag_cal(soft_xz, MagCalRegisterTypes.SOFT, MagCalSoftTypes.XZ)
        self.__write_mag_cal(soft_yz, MagCalRegisterTypes.SOFT, MagCalSoftTypes.YZ)

        # check that the writes were successful
        hard_iron_offset_read, soft_iron_matrix_read = self.read_mag_cal_params()
        equal_hard = np.allclose(
            hard_iron_offset_read,
            hard_iron_offset,
            atol=1 / self.__HARD_IRON_SCALE_FACTOR,
        )
        equal_soft = np.allclose(
            soft_iron_matrix_read,
            soft_iron_matrix,
            atol=1 / self.__SOFT_IRON_SCALE_FACTOR,
        )
        if not equal_hard or not equal_soft:
            raise Exception(
                "Cannot write magnetometer calibration settings - try re-docking Expansion Plate to pi-top [4]."
            )

    def __read_imu_data(self, data_register: int):
        return self.__mcu_device.read_signed_word(data_register, little_endian=True)

    def __get_raw_data(self, data_type: int):
        x = self.__read_imu_data(self.__data_registers[data_type][RawRegisterTypes.X])
        y = self.__read_imu_data(self.__data_registers[data_type][RawRegisterTypes.Y])
        z = self.__read_imu_data(self.__data_registers[data_type][RawRegisterTypes.Z])

        return x, y, z

    def __write_mag_cal(self, word, register_type, calibration_type):
        self.__mcu_device.write_word(
            self.__mag_cal_registers[register_type][calibration_type],
            word,
            little_endian=True,
            signed=True,
        )

    def __read_mag_cal_hard(self):
        data = []
        for register_key, register_value in self.__mag_cal_registers[
            MagCalRegisterTypes.HARD
        ].items():
            data.append(
                self.__mcu_device.read_signed_word(register_value, little_endian=True)
            )
        x, y, z = data
        return x, y, z

    def __read_mag_cal_soft(self):
        data = []
        for register_key, register_value in self.__mag_cal_registers[
            MagCalRegisterTypes.SOFT
        ].items():
            data.append(
                self.__mcu_device.read_signed_word(register_value, little_endian=True)
            )
        xx, yy, zz, xy, xz, yz = data
        return xx, yy, zz, xy, xz, yz



from enum import Enum


class MotorRegisterTypes:
    CONTROL_MODE = 0
    MODE_0_POWER = 1
    MODE_1_RPM = 2
    MODE_2_RPM_WITH_ROTATIONS = 3
    BRAKE_TYPE = 4
    TACHOMETER = 5
    ODOMETER = 6


EncoderMotorM0 = {
    MotorRegisterTypes.CONTROL_MODE: 0x60,
    MotorRegisterTypes.MODE_0_POWER: 0x64,
    MotorRegisterTypes.MODE_1_RPM: 0x68,
    MotorRegisterTypes.MODE_2_RPM_WITH_ROTATIONS: 0x6C,
    MotorRegisterTypes.BRAKE_TYPE: 0x70,
    MotorRegisterTypes.TACHOMETER: 0x75,
    MotorRegisterTypes.ODOMETER: 0x7A,
}


EncoderMotorM1 = {
    MotorRegisterTypes.CONTROL_MODE: 0x61,
    MotorRegisterTypes.MODE_0_POWER: 0x65,
    MotorRegisterTypes.MODE_1_RPM: 0x69,
    MotorRegisterTypes.MODE_2_RPM_WITH_ROTATIONS: 0x6D,
    MotorRegisterTypes.BRAKE_TYPE: 0x71,
    MotorRegisterTypes.TACHOMETER: 0x76,
    MotorRegisterTypes.ODOMETER: 0x7B,
}


EncoderMotorM2 = {
    MotorRegisterTypes.CONTROL_MODE: 0x62,
    MotorRegisterTypes.MODE_0_POWER: 0x66,
    MotorRegisterTypes.MODE_1_RPM: 0x6A,
    MotorRegisterTypes.MODE_2_RPM_WITH_ROTATIONS: 0x6E,
    MotorRegisterTypes.BRAKE_TYPE: 0x72,
    MotorRegisterTypes.TACHOMETER: 0x77,
    MotorRegisterTypes.ODOMETER: 0x7C,
}


EncoderMotorM3 = {
    MotorRegisterTypes.CONTROL_MODE: 0x63,
    MotorRegisterTypes.MODE_0_POWER: 0x67,
    MotorRegisterTypes.MODE_1_RPM: 0x6B,
    MotorRegisterTypes.MODE_2_RPM_WITH_ROTATIONS: 0x6F,
    MotorRegisterTypes.BRAKE_TYPE: 0x73,
    MotorRegisterTypes.TACHOMETER: 0x78,
    MotorRegisterTypes.ODOMETER: 0x7D,
}


class MotorControlRegisters(Enum):
    M0 = EncoderMotorM0
    M1 = EncoderMotorM1
    M2 = EncoderMotorM2
    M3 = EncoderMotorM3


class MotorControlModes(Enum):
    MODE_0 = 0
    MODE_1 = 1
    MODE_2 = 2


class MotorSyncBits(Enum):
    M0 = 0b0000001
    M1 = 0b0000010
    M2 = 0b0000100
    M3 = 0b0001000


class MotorSyncRegisters(Enum):
    CONFIG = 0x57
    START = 0x58



from enum import Enum


class ServoRegisterTypes:
    ACC_MODE = 0
    SPEED = 1
    ANGLE_AND_SPEED = 2


ServoMotorS0 = {
    ServoRegisterTypes.ACC_MODE: 0x50,
    ServoRegisterTypes.SPEED: 0x56,
    ServoRegisterTypes.ANGLE_AND_SPEED: 0x5C,
}

ServoMotorS1 = {
    ServoRegisterTypes.ACC_MODE: 0x51,
    ServoRegisterTypes.SPEED: 0x57,
    ServoRegisterTypes.ANGLE_AND_SPEED: 0x5D,
}

ServoMotorS2 = {
    ServoRegisterTypes.ACC_MODE: 0x52,
    ServoRegisterTypes.SPEED: 0x58,
    ServoRegisterTypes.ANGLE_AND_SPEED: 0x5E,
}

ServoMotorS3 = {
    ServoRegisterTypes.ACC_MODE: 0x53,
    ServoRegisterTypes.SPEED: 0x59,
    ServoRegisterTypes.ANGLE_AND_SPEED: 0x5F,
}


class ServoControlRegisters(Enum):
    S0 = ServoMotorS0
    S1 = ServoMotorS1
    S2 = ServoMotorS2
    S3 = ServoMotorS3


class ServoMotorSetup:
    REGISTER_MIN_PULSE_WIDTH = 0x4A
    REGISTER_MAX_PULSE_WIDTH = 0x4B
    REGISTER_PWM_FREQUENCY = 0x4C



# Foundation Kit
from .button import Button
from .buzzer import Buzzer
from .encoder_motor import EncoderMotor

# Robotics Kit
from .imu import IMU
from .led import LED
from .light_sensor import LightSensor
from .parameters import BrakingType, Direction, ForwardDirection
from .potentiometer import Potentiometer
from .servo_motor import ServoMotor
from .servo_motor import ServoMotorSetting
from .servo_motor import ServoMotorSetting as ServoMotorState
from .sound_sensor import SoundSensor
from .ultrasonic_sensor import UltrasonicSensor
